export type Event_T = [name:string, payload:any];
export interface Machine_T<C> {
  states: Array<State_T<C>>
}
export interface State_T<C> {
  name: string;
  eventReactionCouplings: Array<EventReactionCouplings_T<C>>;
}
export interface EventReactionCouplings_T<C> {
  eventName: string;
  reactions: Array<Reaction_T<C>>;
};
export type Reaction_T<C> = SideEffect_T<C> | ContextMutation_T<C> | Goto_T;
export interface SideEffect_T<C> {
  type: 'SideEffect';
  fn: SideEffectFunction_T<C>;
};
export type SideEffectFunction_T<C> = (ctx:C,e:Event_T,self:Interpreter_T<C>,originalContext:C)=>void;
export interface ContextMutation_T<C> {
  type: 'ContextMutation';
  fn: ContextMutationFunction_T<C>;
};
export type ContextMutationFunction_T<C> = (ctx:C,e:Event_T,self:Interpreter_T<C>)=>C;
export interface Goto_T {
  type: 'Goto';
  targetStateName: string;
};

export const Machine = function<C>(...states:Array<State_T<C>>) : Machine_T<C> { return {states}; };
export const State = function<C>(name:string, ...eventReactionCouplings:Array<EventReactionCouplings_T<C>>) : State_T<C>{ return {name, eventReactionCouplings}; };
export const On = function<C>(eventName:string, ...reactions:Array<Reaction_T<C>>) : EventReactionCouplings_T<C>{ return {eventName, reactions}; };
export const SideEffect = function<C>(fn:SideEffectFunction_T<C>) : SideEffect_T<C>{ return {type:'SideEffect', fn}; };
export const Goto = function(targetStateName:string) : Goto_T { return {type:'Goto', targetStateName} };
export const Context = function<C>(fn:ContextMutationFunction_T<C>) : ContextMutation_T<C> { return {type:'ContextMutation', fn} };





export interface Interpreter_T<C> {
  machine: Machine_T<C>;
  state: string;
  context: C;
  eventQueue:Array<Event_T>;
  subscriptions: Record<string, SubscriptionCallbackFunction_T<C>>;
  isTransitioning: boolean;
  isPaused: boolean;
  start: ()=>Interpreter_T<C>;
  subscribe: (callback:SubscriptionCallbackFunction_T<C>)=>Interpreter_T<C>;
}

/**
 * Description placeholder
 *
 * @export
 * @param {Machine_T} machine
 * @param {any} initialContext
 * @param {?string} [initialStateName]
 * @returns {Interpreter_T}
 */
export function Interpreter<C>(machine:Machine_T<C>, initialContext:any, initialStateName?:string) : Interpreter_T<C>{
  if(typeof initialStateName === 'undefined'){ initialStateName = machine.states[0].name; }
  //@ts-expect-error
  const interpreter : Interpreter_T<C> = {machine, state: initialStateName, context:initialContext, eventQueue:[], isTransitioning:false, subscriptions: {}, isPaused: true};
  interpreter.start = ()=>{ start(interpreter); return interpreter; }
  interpreter.subscribe = (callback)=>{ subscribe(interpreter,callback); return interpreter; }
  send(interpreter, ['entry', null] );
  return interpreter;
}
export function start<C>(interpreter:Interpreter_T<C>){
  if(interpreter.isPaused === true){
    interpreter.isPaused = false;
    processEvents(interpreter);
  }
}
export function pause<C>(interpreter:Interpreter_T<C>){
  if(interpreter.isPaused === false){
    interpreter.isPaused = true;
  }
}

/**  Helper function for `send()`
 */
function getState<C>(interpreter : Interpreter_T<C>) : State_T<C>{
  return interpreter.machine.states.find((state)=>state.name===interpreter.state) as unknown as State_T<C>;
}
/**  Helper function for `send()`
 */
function getMatchingEventReactionCouplings<C>(state : State_T<C>, event:Event_T) : Array<EventReactionCouplings_T<C>>{
  return state.eventReactionCouplings.filter((eventReactionCoupling)=>eventReactionCoupling.eventName===event[0]);
}
/**  Inject an Event into the Interpreter's "tick queue".
 * 
 * An event can be signify something "new" happening, such that its reactions should run on the next Tick;
 * or it can signify a milestone "within" the current Tick, such that a Tick can be thought of as having 
 * "sub-Ticks".
 * 
 * This distinction is significant for proper ordering of reaction execution, and also for determining
 * whether to run a reaction at all. If an Event is received, and is specified to be applied on a past 
 * Tick, it is discarded.
 */
export function send<C>(interpreter : Interpreter_T<C>, event:Event_T){
  interpreter.eventQueue.push(event);
  if(interpreter.isTransitioning === false){
    processEvents(interpreter);
  }
}
export const enqueue = send;
function processEvents<C>(interpreter:Interpreter_T<C>){
  interpreter.isTransitioning = true;
  while(interpreter.eventQueue.length > 0 && interpreter.isPaused===false){
    processNextEvent(interpreter);
  }
  interpreter.isTransitioning = false;
  // only run subscriptions here, once the machine's state has settled:
  Object.values(interpreter.subscriptions).forEach((subscriptionCallbackFunction)=>{ subscriptionCallbackFunction(interpreter); });
}
function processNextEvent<C>(interpreter:Interpreter_T<C>){
  const event = interpreter.eventQueue.shift();
  if(typeof event !== 'undefined'){
    const state = getState(interpreter);
    const eventReactionCouplings = getMatchingEventReactionCouplings(state, event);
    const reactions = eventReactionCouplings
      .map((eventReactionCoupling)=>eventReactionCoupling.reactions)
      .flat();
    const {sideEffects, contextMutations, goto_} = categorizeReactions(reactions);
    // save the current context, before it's mutated, so as to pass it to sideEffects below:
    const originalContext = interpreter.context;
    // must process contextMutations in-series:
    contextMutations.forEach((contextMutation)=>{
      interpreter.context = contextMutation.fn(interpreter.context, event, interpreter);
    });
    // can process sideEffects in parallel (though we currently don't due to the overhead of doing so in Node.js):
    // they're processed *after* the context changes, since that's what most sideEffects would be interested in; but nevertheless the original context is passed in case this sideEffect needs it:
    sideEffects.forEach((sideEffect)=>{
      sideEffect.fn(interpreter.context, event, interpreter, originalContext);
    });
    // processing of `goto` must be last:
    if(goto_ !== null){
      send(interpreter, ['exit', null]);
      interpreter.state = goto_.targetStateName;
      send(interpreter, ['entry', null]);
    }
  }
}
function categorizeReactions<C>(reactions:Array<Reaction_T<C>>) : {sideEffects:Array<SideEffect_T<C>>, contextMutations:Array<ContextMutation_T<C>>, goto_:Goto_T|null}{
  let 
    sideEffects:Array<SideEffect_T<C>> = [], 
    contextMutations:Array<ContextMutation_T<C>> = [], 
    goto_:Goto_T|null = null;
  reactions.forEach((reaction)=>{
    if(reaction.type === 'SideEffect'){
      sideEffects.push(reaction);
    }
    else if(reaction.type === 'ContextMutation'){
      contextMutations.push(reaction);
    }
    else if(reaction.type === 'Goto'){
      goto_ = reaction;
    }
  });
  return {sideEffects, contextMutations, goto_};
}

export type SubscriptionCallbackFunction_T<C> = (self:Interpreter_T<C>)=>void;
let subscriptionId : number = 0;
export function subscribe<C>(interpreter:Interpreter_T<C>, callback:SubscriptionCallbackFunction_T<C>){
  subscriptionId++;
  interpreter.subscriptions[subscriptionId.toString()] = callback;
  return subscriptionId.toString();
}
export function unsubscribe<C>(interpreter:Interpreter_T<C>, subscriptionId:string){
  delete interpreter.subscriptions[subscriptionId.toString()];
}

export const Spawn = function(){};
export const Unspawn = function(){};

/*
export function useMachine(machine, options){
  return useMemo(()=>interpret(AppMachine, {context:{}}),[]);
}
*/